﻿/* ir_rx.c : 赤外線リモコン受信 サンプルプログラム
 *
 * コンパイル例:
 *	gcc -o ir_rx ir_rx.c -lrt
 *  (clock_gettime関数を使用するためリアルタイム拡張オプションが必要)
 *
 * 実行時オプション設定(引数設定)：
 *	sudo ./ir_rx <時間間隔を格納するファイル名> で実行. 受信した時間間隔ファイルが生成されます.
 *
 * エアコンのリモコンを[冷房]/[最低温度]に設定して,
 * 電源ON  -> ac_on.dat
 * 電源OFF -> ac_off.dat のファイル名で記録(学習)してください.
 *
 * 受信時は、リモコンを受光部に近づけて(正面を向け)学習させるのがコツです.
 * 何回か試して良好な送信結果が得られるよう実験してください.
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <ctype.h>
#include <fcntl.h>
#include <poll.h>
#include <time.h>
#include <errno.h>
#include <getopt.h>
#include <termios.h>
#include <unistd.h>

#define PORT_NUM 18
#define MAX_PULSE 2048 				//最大受信パルス数
#define MAX_USEC 300000				//最長受信時間(usec)
#define RX_TIMEOUT 2000 			//受信タイムアウト値(msec)


int main(int argc,char *argv[]){

	int port = PORT_NUM;			//赤外線受信モジュールを接続したポート番号
	int fd;							//GPIOアクセス用 openで開く(バッファを使用しない)
	int ret;

	char buf[40];					//sprintf用
	char val[8];					//gpioをreadした値
	
	long temp;
	long total_wait;				//受信開始からのトータル時間
	long wait[MAX_PULSE*2];			//受信データの時間間隔を格納する配列
	int cnt;						//時間間隔データのカウント

	
	struct timespec nowts,prets;	//clock_gettime関数で取得した時間を格納
	struct pollfd pfd;				//poll関数で使用するディスクリプタ

	//時間間隔ファイルオープン用
	FILE *fp;						//ファイルはfopenで開く(バッファ使用)
	char *filename; 
	int j;


	//引数チェック
	if (argc != 2 ) {
		fprintf(stderr, "引数が正しくありません．(受信データ格納ファイルを指定)\n");
		return -1;
	}
	filename = argv[1];				//引数をファイル名に

	//ポート初期化
	//--ここではポート番号(変数)をsprintf文で文字列に変換→書きこむ方法で処理
	fd = open("/sys/class/gpio/export", O_WRONLY);
	sprintf(buf, "%d", port);
	write(fd, buf, strlen(buf));
	close(fd);

	//入力に設定
	sprintf(buf, "/sys/class/gpio/gpio%d/direction", port);
	fd = open(buf, O_WRONLY);
	write(fd, "in", 2);
	close(fd);

	//割り込みイベント設定
	sprintf(buf, "/sys/class/gpio/gpio%d/edge", port);
	fd = open(buf, O_WRONLY);
	write(fd, "both", 4);
	close(fd);

	//入力でvalue(ポート)をオープン
	sprintf(buf, "/sys/class/gpio/gpio%d/value", port);
	fd = open(buf, O_RDONLY);
printf("FOPEN:OK\n");


//--------赤外線データ受信ルーティン

	pfd.fd = fd; 					//ファイルディスクリプタ設定
	pfd.events = POLLPRI; 			//イベント種別設定 値変更で発生
	read(fd, val, 1);				//ダミーリードしておく

//--------マクロ(ルーティン)定義
/*poll関数 イベント監視ループのマクロ
  変数 ret:poll関数の戻り値(エッジ検出成否) / val:入力ポートの値
  エッジ変化直後にclock_gettime関数で現在時間を取得しておく */
#define poll_edge(timeout) \
	lseek(fd, 0, SEEK_SET); \
	ret = poll(&pfd, 1, timeout); \
	clock_gettime(CLOCK_MONOTONIC_RAW, &nowts); \
	read(fd, val, 1)


/*前回計測時からの差分を計算→記録するマクロ
  wait[]にデータのON⇔OFF間隔を記録 */
#define rec_time() \
	temp =  (nowts.tv_sec - prets.tv_sec) * 1000000; \
	temp += (nowts.tv_nsec - prets.tv_nsec) / 1000; \
	wait[cnt++] = temp; \
	total_wait += temp; \
	prets = nowts


//--------赤外線受信開始
	do{
		//はじめに立ち上がりを検出待ち (リーダー部の先頭)
		poll_edge(RX_TIMEOUT);
		if(ret <= 0){
			printf("受信タイムアウト!\n");
			exit(-1);
		}
	//printf("P1\n");
		} while (val[0] == '1');	//立ち上がり検出まで繰り返し

	prets = nowts;					//現在時間→前の時間にしておく

	cnt = 0;
	total_wait = 0;

	// signal record
	while(1){
		while( cnt < MAX_PULSE ){
			//立下り検出待ち
			poll_edge(RX_TIMEOUT);
			if(ret <= 0){
				printf("タイムアウト!(受信失敗)\n");
				exit(-1);
			}
			rec_time();				//立ち上がりからの時間の差分を記録

			//立ち上がり検出待ち その1
			poll_edge(30);
			if(ret <= 0 ) break; 	// >30msでタイムアウト.一旦ループを抜ける
			rec_time();				//立ち下がりからの時間の差分を記録
		}
		//レコード長：MAX_USEC以上で受信終了
  		if (total_wait > MAX_USEC){
			rec_time();				//OFF時間を記録しておく(ダミー)
			break;
		}
		//立ちあがり検出待ち その2
		poll_edge(70);
		rec_time();					//OFF時間を記録しておく
		if(ret <= 0) break;			// >70msでタイムアウト.都合100msで受信終了
									// 都合100ms以内なら受信続行
		//printf("P5\n");
	} 

	//valueポート(ファイル)を閉じる
	close(fd);
	// I/O解放
	fd = open("/sys/class/gpio/unexport", O_WRONLY);
	sprintf(buf, "%d", port);
	write(fd, buf, strlen(buf));
	close(fd);


//--------時間間隔をファイルに保存

	//ファイルオープン
	if ((fp = fopen(filename, "w")) == NULL) {
		fprintf(stderr, "%sのオープンに失敗しました．\n", filename);
		exit(-1);
	}

	fprintf(fp, "%d\n",cnt);
	for(j=0; j<cnt; j=j+2){
    	fprintf(fp, "%ld,%ld\n", wait[j],wait[j+1]);
	}
	fclose(fp);		//ファイルクローズ
	printf("受信完了!\n");

}//main

