/* mdint.c : モーションセンサ入力 簡易割り込み制御 サンプルプログラム
 *
 * コンパイル例:
 *	gcc -o mdint mdint.c
 *
 * 実行例
 *	sudo ./mdint [-d/-k/null]
 *
 * 実行時オプション設定(引数設定)：
 *		-d 	  : デーモン化
 *		-k	  : kill(デーモンプロセスを停止)
 *		(なし): daemon化せず起動
 *
 *	※実行時、ファイル:[pid_file_path]を生成し、
 *	  起動したデーモンプロセスのPIDを記録します.
 *	  (後でプロセスをkillする時に使用する)
 *	※デーモン化したプロセスは [ps aux]コマンドなどで確認できます.
 */

#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>
#include <syslog.h>
#include <poll.h>
#include <sys/signal.h>
#include <fcntl.h>
#include <time.h>

#define FLAG_DAEMONIZE     1
#define FLAG_KILL          2

#define PORT_NUM 27		//ポート番号(マクロ)
#define OFF_LONG 10000	//poll関数タイムアウト値(ms)
#define MAX_REC  100	//時間レコードバッファの最大値
#define DET_TIME 600	//動き検出の基準時間(sec) (この時間内に)
#define DET_CNT	 30		//動き検出の基準回数      (この回数動いたら)
#define DET_NMD	 1		//フラグ設定 =動きなし(寝た)と検出
#define DET_AMD	 0		//フラグ設定 =動きあり(起きてる)と検出(デフォルト)

#define pid_file_path "mdint.pid"	//プロセスIDを記録するファイル名

int main(int argc, char *argv[]){

	//daemon化用変数
	int flag = 0;		//daemon動作用フラグ
	int nochdir = 1;	//0以外:カレントディレクトリを "/" (ルート) にしない
	int noclose = 0;	//0    :標準入力/標準出力/標準エラーをクローズ
	FILE *pid_fd;		//プロセスID書き出し用 ファイルディスクリプタ
	pid_t pid;			//プロセスID用変数

	//モーションセンサ割り込み制御用変数
	int i,j;
	int fd;
	int ret;
	int rec_cnt,nmd_flag;
	struct pollfd pfd;
	char val[8];
	time_t now_time,rec_time[100];


	if (argc == 2 ) {			//引数チェック
		if(*argv[1] == '-'){
			switch(*(argv[1] + 1)){
				case 'd':
					flag = FLAG_DAEMONIZE;
					//printf("daemonize\n");
					break;
				case 'k':
					flag = FLAG_KILL;
                	//printf("kill\n");
					break;
				default:
					printf("OPTION: [-d]:daemonize / [-k]:kill\n");
					return -1;
			}
		}
	}
	//このオプション以外は通常実行

	//[/var/log/]にdaemon[PID]:としてログを出力
	openlog("daemon", LOG_PID, LOG_DAEMON);


	//オプション[-k]: 保存したpidを用いてデーモンプロセスを終了する
	if (flag == FLAG_KILL) {
		pid_fd = fopen(pid_file_path, "r");
		if (pid_fd != NULL) {
			fscanf(pid_fd, "%d\n", &pid);
			fclose(pid_fd);
			if (kill(pid, SIGKILL) == 0) {
				syslog(LOG_INFO, "daemon stopped.\n");
			}
		} else {
			syslog(LOG_ERR, "no daemon started.\n");
			return -1;
		}
		return 0;
	}

	//オプション[-d]: デーモンプロセスを立ち上げる
	if (flag == FLAG_DAEMONIZE) {
		if (daemon(nochdir, noclose) == -1) {
			syslog(LOG_ERR, "failed to launch daemon.\n");
			return -1;
		}
	}

	syslog(LOG_INFO, "daemon started.\n");

	// 現プロセス(デーモン)のPIDを取得して ファイルに保存(記録しておく)
	// ※あとでkillする時に必要
	pid = getpid();
	pid_fd = fopen(pid_file_path, "w+");
	if (pid_fd != NULL) {
		fprintf(pid_fd, "%d\n", pid);
		fclose(pid_fd);
	} else {
		syslog(LOG_ERR, "failed to record process id to file.\n");
		return -1;
	}

//--------------- ここからモーションセンサ入力の割り込み処理(daemon化する部分)

	now_time = time((time_t *)0);	//現在時間を取得
	rec_cnt = 0;					//時間レコードバッファ・カウンタ初期化
	for (i=0;i<MAX_REC;i++){
		rec_time[i] = now_time;
	}
	nmd_flag = DET_AMD;				//動きなし検出フラグ初期化


	//(モーションセンサを接続した)GPIOポートの入力→割り込みエッジ設定

	//ポート初期化 27番ポートをモーションセンサ入力に使用
	fd = open("/sys/class/gpio/export", O_WRONLY);
	write(fd, "27", 2);
	close(fd);

	//入力に設定
	fd = open("/sys/class/gpio/gpio27/direction", O_WRONLY);
	write(fd, "in", 2);
	close(fd);

	//割り込みイベント設定
	fd = open("/sys/class/gpio/gpio27/edge", O_WRONLY);
	write(fd, "falling", 7);
	close(fd);

	//入力でvalue(ポート)をオープン
	fd = open("/sys/class/gpio/gpio27/value", O_RDONLY);
	pfd.fd = fd;
	pfd.events = POLLPRI;
	read(fd, val ,sizeof(val));			//ダミーリードしておく


	//モーションセンサ割り込み イベント監視ループ
	while(1){
		//Falling Edge : 1 -> 0 変化待ち
		lseek(fd, 0, SEEK_SET);			// /sys/～/gpio/valueレコード巻き戻し
		ret = poll(&pfd, 1, OFF_LONG);	// 変化待ち OFF_LONGでタイムアウト
		read(fd, val, sizeof(val));		// 値を取得
		if (ret>0){						//割り込み=センサ感あり。現在時間をリングバッファに記録
			rec_time[rec_cnt++] = time((time_t *)0);
			if (rec_cnt >= MAX_REC) rec_cnt = 0;
		}

		i = 0;
		j = rec_cnt;
		now_time = time((time_t *)0);	//現在時刻を更新

		while(1){						//リングバッファを検索。基準時間内の反応回数を計測
			if (--j<0) j = MAX_REC - 1;
			if ((difftime(now_time,rec_time[j])) > DET_TIME) break;
			if (++i >= MAX_REC -1 ) break;

		}
		if (i < DET_CNT){				//基準値以下->動きなしと判定
			if (nmd_flag == DET_AMD){	//main.pyを呼び出し→パラメータ変更(1回のみ実行)
				nmd_flag = DET_NMD;
				if(system("/home/pi/work/mainrht.py NMD_ON")== -1){
					printf("SYSERR");
				}
				//printf("low\n");
			}
		}else{
			if (nmd_flag == DET_NMD){	//基準値以上->動きあり
				nmd_flag = DET_AMD;		//main.pyを呼び出し→パラメータ変更(1回のみ実行)
				if(system("/home/pi/work/mainrht.py NMD_OFF")== -1){
					printf("SYSERR");
				}
				//printf("high\n");
			}
		}
	}
	close(fd);	// value(ポート)クローズ
	
	// I/O解放
	fd = open("/sys/class/gpio/unexport", O_WRONLY);
	write(fd, "27", 2);
	close(fd);

	return(0);
}
